// randchirp.cc -- generate a regular random chirp

// To use this file, your version of Octave must support dynamic
// linking.  To find out if it does, type the command
//
//   x = octave_config_info; x.DEFS
//
// at the Octave prompt.  Support for dynamic linking is included if
// the output contains the string -DWITH_DYNAMIC_LINKING=1.
//
// To compile this file, type the command
//
// mkoctfile --verbose randchirp.cc
//
// at the shell prompt.  The script mkoctfile should have been
// installed along with Octave.  Running it will create a file called
// randchirp.oct that can be loaded by Octave.
//
// Enter:
//
// help randchirp
//
// at the Octave prompt to have a description of how randchirp should be used.  
// 
// This file may be used under the terms defined by the GNU 
// General Public License [write to the Free Software Foundation, 
// 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA]
//
// WITHOUT ANY WARRANTY; without even the implied warranty 
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// Copyright (C) 2005 Eric Chassande-Mottin, CNRS (France)

//

#include <octave/config.h>

#include <iostream>
#include <sstream>
using namespace std;
#include <string>

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

# define RANDMAX 4096
#define M_TWOPI     6.2831853071795864769252867665590058L  /* 2*pi */

int debug_flag=0;

static inline int
getmin ( int x, int y)
{
  return x < y ? x : y;
}
static inline double getrand(void)
{ 
  return (rand() % RANDMAX)/((double) RANDMAX);
}

DEFUN_DLD (randchirp, args, ,
  "RANDCHIRP:  generates a regular random chirp\n")
{
  octave_value_list retval;
  
  /*----------- get input parameters ----------------*/
  int nargin = args.length ();
  
  if ((nargin < 5)|(nargin > 8))
    {
      error("Usage: out = randchirp(N,b,Nr1,step,Nr2[,ml,mu,shift])");
      return retval;
    }

  /* get the data */
  int N=args(0).int_value();

  /* check that c is a power of 2 */
  while(!(N & 1))
      N = N>>1;

  if (!(N==1))
    {
      error("randchirp: N should be a power of 2");
      return retval;
    }
  N=args(0).int_value();

  /* get the size of the interval */
  int b =args(1).int_value();

  /* check that b is a power of 2 */
  while(!(b & 1))
      b = b>>1;

  if (!(b==1))
    {
      error("randchirp: b should be a power of 2");
      return retval;
    }
  b=args(1).int_value();

  if ((b<2)|(b>=N))
    {
      error("randchirp: b should be 0 < b < N");
      return retval;
    }
  int Nt=(int) N/b;

  /* get the number of steps, constraint 1: 1st derivative */
  int Nr1 =args(2).int_value();

  /* get the step size*/
  int step =args(3).int_value();

  /* get the number of steps, constraint 2: 2nd derivative */
  int Nr2 =args(4).int_value();
  if (Nr2>=(2*Nr1))
      printf("randchirp: warning! no effective constraint on 2nd derivative");

  /* get the boundaries of the frequency axis */
  int ml=0;
  int mu=N;
  if (nargin>=6)
    ml=args(5).int_value();
  if (nargin>=7)
    mu=args(6).int_value();

  if (ml>mu)
    error("randchirp:  we should have ml <= mu");
  int Nf=mu-ml+1;

  /* shift the frequency axis ? */
  double shift_f=0.0;
  if (nargin==8)
    if (args(7).int_value()==1)
      shift_f=0.5;

  ColumnVector f(Nf);
  for (int k=ml; k<=mu; k++)
    f(k-ml)=(double) k/(2.0*(double) N)+shift_f;

  /* first chirplet */
  ColumnVector m(Nt+1); //int Array ?
  m(0)=round((Nf-1)*getrand());

  int n=0,dn=0,mn=0;
  int dl=-Nr1, du=Nr1;

  // make steps ahead and correct upper bound if necessary
  do
    {
      n=1; 
      dn=du;
      mn=(int) m(0)+dn;
      while ((n<=Nt)&(mn<Nf)&(dn>0))
	{
	  dn-=Nr2;
	  mn+=dn;
	  n++;
	}

      if (mn>=Nf)
	du-=step; 
      else
	break;
    }
  while (1);


  // make steps ahead and correct lower bound if necessary
  do
    {
      n=1; 
      dn=dl;
      mn=(int) m(0)+dn;
      while ((n<=Nt)&(mn>=0)&(dn<0))
	{
	  dn+=Nr2;
	  mn+=dn;
	  n++;
	}
      
      if (mn<0)
	dl+=step;
      else
	break;
    }
  while (1);
  int d=(int) round(dl+(du-dl)*getrand());
  m(1)=m(0)+d;

  /* main loop */
  for (int j=2; j<Nt; j++)
    {
      // define bounds for chirplet variation
      dl=-getmin((int) m(j-1),getmin(Nr1,Nr2-d));
      du=getmin(Nf-1-(int) m(j-1),getmin(Nr1,Nr2+d));
      if (debug_flag==1)
	{
	  printf("chirplet %d\n",j);
	  printf("st: dl=%d du=%d\n",dl,du);  fflush(stdout);
	}

      // make steps ahead and correct upper bound if necessary
      if (du>0)
	{
	  do
	    {
	      n=j+1; 
	      dn=du;
	      mn=(int) m(j-1)+dn;
	      while ((n<=Nt)&(mn<Nf)&(dn>0))
		{
		  dn-=Nr2;
		  mn+=dn;
		  n++;
		}

	      if (debug_flag==1)
		printf("n=%d mn=%d dn=%d\n",n,mn,dn);
	      if (mn>=Nf)
		du-=step; 
	      else
		break;
	    }
	  while (1);
	}

      if (debug_flag==1)
	{printf("up: dl=%d du=%d\n",dl,du); fflush(stdout);}

      // make steps ahead and correct lower bound if necessary
      if (dl<0)
	{
	  do
	    {
	      n=j+1; 
	      dn=dl;
	      mn=(int) m(j-1)+dn;
	      while ((n<=Nt)&(mn>=0)&(dn<0))
		{
		  dn+=Nr2;
		  mn+=dn;
		  n++;
		}

	      if (debug_flag==1)
		printf("n=%d mn=%d dn=%d\n",n,mn,dn);
	      if (mn<0)
		dl+=step;
	      else
		break;
	    }
	  while (1);
	}

      if (debug_flag==1)
	{printf("lo: dl=%d du=%d\n",dl,du); fflush(stdout);}

      if (debug_flag==1)
	{printf("fi: dl=%d du=%d\n",dl,du); fflush(stdout);}

      d=(int) round(dl+(du-dl)*getrand());
      m(j)=m(j-1)+d;

      if (debug_flag==1)
	printf("m(j-1)=%d m(j)=%d d=%d\n\n",(int) m(j-1),(int) m(j),d);
    }
  
  dl=-getmin((int) m(Nt-1),getmin(Nr1,Nr2-d));
  du= getmin(Nf-1-(int) m(Nt-1),getmin(Nr1,Nr2+d));
  d=(int) round(dl+(du-dl)*getrand());
  m(Nt)=m(Nt-1)+d;

  //   int mmax=(int) m(0), mmin=(int) m(0);
  //   for (int j=1 ; j<=Nt ; j++)
  //     {
  //       if (m(j)>mmax)
  // 	mmax=(int) m(j);
  //       if (m(j)<mmin)
  // 	mmin=(int) m(j);
  //     }
  //printf("min=%d max=%d\n",mmin,mmax);
  //   if ((mmin<0)|(mmax>=Nf))
  //     printf("min=%d max=%d\n",mmin,mmax);

  /* create signal */
  ColumnVector out(N);
  out.fill(0.0);

  /* create frequency */
  ColumnVector ifl(N);
  ifl.fill(0.0);

  /* main loop */
  double a2,a1,b2,b1;
  double theta=0.0;

  int i=0;
  for (int j=0 ; j<Nt ; j++)
    {
      b2=(f((int) m(j+1))-f((int) m(j)))/((double) b);
      b1=f((int) m(j));
      a2=M_PI*b2;
      a1=M_TWOPI*f((int) m(j));
      out(i)=theta;
      ifl(i)=b1; i++;
      n=1;
      while (n<b)
        {
          out(i)=a2*n*n+a1*n+theta;
          ifl(i)=b2*n+b1;
          n++;i++;
        }
      theta+=M_PI*((double) b)*(f((int) m(j+1))+f((int) m(j)));
    }

  retval(0)=out;
  retval(1)=ifl;
  retval(2)=m;

  return retval;
}
