// tfc.cc -- interface to tfcluster algorithm.

// To use this file, your version of Octave must support dynamic
// linking.  To find out if it does, type the command
//
//   x = octave_config_info; x.DEFS
//
// at the Octave prompt.  Support for dynamic linking is included if
// the output contains the string -DWITH_DYNAMIC_LINKING=1.
//
// You need to have the LAL (Ligo Algorithm Library, www.ligo.org) 
// to compile this code.
//
// Enter:
//
// help tfc
//
// at the Octave prompt to have a description of how sts should be used.  
// 
// This file may be used under the terms defined by the GNU 
// General Public License [write to the Free Software Foundation, 
// 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA]
//
// WITHOUT ANY WARRANTY; without even the implied warranty 
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// Copyright (C) 2005 Eric Chassande-Mottin, CNRS (France)


//

#include <octave/config.h>

#include <iostream>
#include <sstream>
using namespace std;
#include <string>

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

#include <lal/AVFactories.h>
#include <lal/TFClusters.h>

DEFUN_DLD (tfc, args, ,
  "TFC: tfcluster algorithm.\n\
Usage:\n\
[ptot,N,t,f,p,idx] = tfc(tfr[,thr,alpha])\n\
Input parameters:\n\
tfr : time-frequency matrix (spectrogram)\n\
thr : defines threshold on tfr such that 1 every thr pix is black (default =10)\n\
alpha : select (a prop of) alpha of all clust from white noise will make it (default =.25)\n\
Output parameters:\n\
ptot : integrated tfr for each clusters\n\
N : number of clusters\n\
t,f,p : coordinates and tfr value for each cluster points\n\
idx : start index of each clusters (in t,f and p).")
{  
  octave_value_list retval;

  /*----------- get input parameters ----------------*/
  int nargin = args.length ();
  
  if (nargin < 1)
    {
      // print_usage("sts");
      error("Usage: [ptot,N,x,y,p,idx] = tfc(x[,thr,alpha])");
      return retval;
    }

  static LALStatus status;

  /* get the data */
  Matrix tfr=args(0).matrix_value();
  int r=tfr.rows();
  int c=tfr.columns();

  double thr=10;
  if (nargin>1)
    thr=args(1).double_value(); //set thr on power so that 1 every thr pixel is black

  double alpha=.25;
  if (nargin>2)
    alpha=args(2).double_value(); //only 1/alpha of all clust from white noise will make it
  
  /* convert to spower format */
  Spectrogram spower;
  spower.power = NULL;
  spower.power=(REAL8 *)malloc(r*c*sizeof(REAL8));
  spower.params = NULL;
  spower.params=(TFPlaneParams *)malloc(sizeof(TFPlaneParams));
  for (int i=0;i<r;i++)
    for (int j=0;j<c;j++)
      spower.power[j*r+i]=tfr(i,j);

  /* create params struc */
  spower.params->freqBins=r; 
  spower.params->timeBins=c;
  spower.params->deltaT=1.0;
  spower.params->flow=0.0;

  /* Set thresholds */
  CListDir dir;
  dir.freqBins = r; /* number of frequency bins in spectrogram */
  dir.sigma = 5; /* threshold on cluster size */
  dir.minf = 0;
  dir.maxf = 512; /* max frequency to consider (Hz) */

  LALFillCListDir(&status, &dir, -log(1/thr));

  /* set thresholds on distance for different size couples */
  dir.d[0] = 0; /* 1,1 */
  dir.d[1] = 0; /* ... */
  dir.d[2] = 0;
  dir.d[3] = 0; /* 1,4 */
  dir.d[4] = 0; /* 2,2 */
  dir.d[5] = 0; 
  dir.d[6] = 2; /* 2,4 */
  dir.d[7] = 3; /* 3,3 */
  dir.d[8] = 4; /* 3,4 */
  dir.d[9] = 4; /* 4,4 */
  dir.mdist = 4;

  /* run cluster threshold algorithm */
  CList clist, list;
  LALInitCList(&status, &clist, spower.params);
  LALGetClusters(&status, &clist, &spower, &dir); 

  /* run threshold on cluster total power */
  dir.alpha = alpha;
  LALInitCList(&status, &list, spower.params); /* initialize list */
  LALClustersPowerThreshold(&status, &list, &clist, &dir); /* generate new list */

  /* clean up a bit */
  LALFreeCList(&status, &clist);
  LALFreeCListDir(&status, &dir);
  
  /* at this point, list is the list of significant clusters */
  int N=list.nclusters;
  int ltot=0;
  if (N>0)
    {
      ColumnVector idx (N);      
      for(int i=0;i<N;i++)
	{
	  idx(i)=ltot+1;
	  ltot+=list.sizes[i];
	}

	ColumnVector x (ltot);
	ColumnVector y (ltot);
	ColumnVector p (ltot);
	ColumnVector ptot (N);

	/* compute the max power among significant clusters */
	int size_max=0;
	double pmax=0.0;
	for(int i=0;i<N;i++)
	  {
	    double p_tot=0.0;
	    for(int j=0;j<list.sizes[i];j++)
	      {
	      x((int) idx(i)-1+j)=list.t[i][j]+1;
	      y((int) idx(i)-1+j)=list.f[i][j]+1;
	      p((int) idx(i)-1+j)=list.P[i][j];	  
	      p_tot+=list.P[i][j];
	    }
	    ptot(i)=p_tot;
	    if (p_tot>pmax)
	      pmax=p_tot;	 
	  }

	retval(0)=pmax;
	retval(1)=ptot;
	retval(2)=(double) N;
	retval(3)=x;
	retval(4)=y;
	retval(5)=p;
	retval(6)=idx;
    }
  else
    {
      retval(0)=0.0;
      retval(1)=0.0;
      retval(2)=0.0;
      retval(3)=0.0;
      retval(4)=0.0;
      retval(5)=0.0;
      retval(6)=0.0;
    }
  /* clean up */
  LALFreeCList(&status, &list);

  return retval;
}






